"use strict";

exports.isFuture = isFuture;
exports.Scheduler = Scheduler;
exports.Future = Future;
exports.isDeinitError = isDeinitError;
function isFuture(e) {
    return isObject(e) && isFunction(e.settle) && isFunction(e.map) && isFunction(e.deinit);
};

function Queue() {
    var e = this;
    validateInstance(e, Queue);
    e.length = 0;
    e.e = [];
    e.t = [];
    e.i = 0;
    e.n = 0;
}

Queue.prototype = {
    constructor: Queue,
    push: function (e) {
        this.length++;
        this.t.push(e);
    },
    shift: function () {
        var e = this;
        if (e.i >= e.n) {
            var t = e.e;
            t.length = 0;
            e.e = e.t;
            e.t = t;
            e.i = 0;
            e.n = e.e.length;
            if (!e.n) return undefined;
        }
        var i = e.e[e.i];
        e.e[e.i++] = undefined;
        e.length--;
        return i;
    }
};

function Scheduler(e) {
    var t = this;
    validateInstance(t, Scheduler);
    validate(e, isFunction);
    t.u = new Queue();
    t.r = e;
    t.s = false;
    t.o = scheduledTick.bind(null, t);
    t.asap = chooseAsapImplementation(t.o);
};

Scheduler.prototype = {
    constructor: Scheduler,
    push: function (e) {
        this.u.push(e);
        if (this.u.length) scheduleTick(this);
    },
    tick: function () {
        var e = this.r;
        try {
            while (this.u.length) {
                e(this.u.shift());
            }
        } finally {
            if (this.u.length) scheduleTick(this);
        }
    },
    deinit: function () {
        while (this.u.length) {
            this.u.shift();
        }
    }
};

function scheduleTick(e) {
    if (!e.s) {
        e.s = true;
        e.asap.call(undefined, e.o);
    }
}

function scheduledTick(e) {
    e.s = false;
    e.tick();
}

var PENDING = 1 | 0;

var ERROR = 2 | 0;

var SUCCESS = 4 | 0;

var PENDING_REJECTION = 8 | 0;

var CONSUMED = 16 | 0;

var MAPPING = 32 | 0;

var SET_ALL = 255 | 0;

var UNSET_STATE = SET_ALL ^ (PENDING | ERROR | SUCCESS);

var SETTLED = ERROR | SUCCESS;

function Future() {
    var e = this;
    validateInstance(e, Future);
    e.f = PENDING;
    e.c = undefined;
    e.a = undefined;
    e.l = undefined;
    e.d = undefined;
    e.h = undefined;
    e.S = undefined;
};

Future.prototype = {
    constructor: Future,
    get state() {
        var e = this;
        return {
            PENDING: someBitsSet(e, PENDING),
            ERROR: someBitsSet(e, ERROR),
            SUCCESS: someBitsSet(e, SUCCESS),
            PENDING_REJECTION: someBitsSet(e, PENDING_REJECTION),
            CONSUMED: someBitsSet(e, CONSUMED),
            MAPPING: someBitsSet(e, MAPPING)
        };
    },
    deref: function () {
        if (someBitsSet(this, ERROR)) {
            unsetBits(this, PENDING_REJECTION);
            throw this.c;
        }
        return this.c;
    },
    settle: function (e, t) {
        var i = this;
        if (someBitsSet(i, SETTLED | MAPPING)) return;
        if (e === i || t === i) {
            throw Error("A future can't be chained to itself");
        }
        if (e) t = undefined;else e = undefined;
        if (isFuture(e)) {
            if (isBaseFuture(e)) {
                if (someBitsSet(e, CONSUMED)) throw Error(CONSUMABLE_ERROR);
                if (!e.l && !i.h) {
                    i.h = alwaysThrow;
                    setupPredecessorSuccessorPair(e, i);
                    return;
                }
            }
            i.a = e.map(function errorSettle(e, t) {
                i.settle(e || t);
            });
            return;
        }
        if (isFuture(t)) {
            if (isBaseFuture(t)) {
                if (someBitsSet(t, CONSUMED)) throw Error(CONSUMABLE_ERROR);
                if (!t.l) {
                    setupPredecessorSuccessorPair(t, i);
                    return;
                }
            }
            i.a = t.map(function valueSettle(e, t) {
                i.settle(e, t);
            });
            return;
        }
        var n = i.h;
        i.h = undefined;
        if (n) {
            setBits(i, MAPPING);
            try {
                t = n(e, t);
                e = undefined;
            } catch (i) {
                e = i;
                t = undefined;
            } finally {
                unsetBits(i, MAPPING);
            }
            i.settle(e, t);
            return;
        }
        i.c = e || t;
        replaceStateBits(i, e ? ERROR : SUCCESS);
        if (e) setBits(i, PENDING_REJECTION);
        scheduleFuture(i);
    },
    map: function (e) {
        validate(e, isFunction);
        if (someBitsSet(this, CONSUMED)) throw Error(CONSUMABLE_ERROR);
        setBits(this, CONSUMED);
        unsetBits(this, PENDING_REJECTION);
        var t = new Future();
        t.h = e;
        setupPredecessorSuccessorPair(this, t);
        return t;
    },
    mapError: function (e) {
        validate(e, isFunction);
        return this.map(mapError.bind(null, e));
    },
    mapResult: function (e) {
        validate(e, isFunction);
        return this.map(mapResult.bind(null, e));
    },
    finally: function (e) {
        validate(e, isFunction);
        return this.map(mapFinally.bind(null, e));
    },
    weak: function () {
        var e = new Future();
        if (!this.d) this.d = new Queue();
        this.d.push(e);
        scheduleFuture(this);
        return e;
    },
    toPromise: function () {
        var e = this;
        if (someBitsSet(e, CONSUMED)) throw Error(CONSUMABLE_ERROR);
        unsetBits(e, PENDING_REJECTION);
        if (someBitsSet(e, ERROR)) return Promise.reject(e.c);
        if (someBitsSet(e, SUCCESS)) return Promise.resolve(e.c);
        return new Promise(mapFutureToPromise.bind(null, e));
    },
    catch: function (e) {
        return this.toPromise().catch(e);
    },
    then: function (e, t) {
        return this.toPromise().then(e, t);
    },
    finishPending: function () {
        var e = this;
        if (someBitsSet(e, SETTLED)) {
            var t = e.l;
            e.l = undefined;
            if (t) {
                unsetBits(e, PENDING_REJECTION);
                if (someBitsSet(e, ERROR)) t.settle(e.c);else t.settle(undefined, e.c);
            }
            while (e.d && e.d.length) {
                var i = e.d.shift();
                if (someBitsSet(e, ERROR)) i.settle(e.c);else i.settle(undefined, e.c);
            }
            finalize(e);
        }
        if (someBitsSet(e, PENDING_REJECTION)) {
            unsetBits(e, PENDING_REJECTION);
            e.constructor.onUnhandledRejection(e);
        }
    },
    deinit: function () {
        var e = this;
        if (someBitsSet(e, MAPPING)) return;
        var t = e.a;
        e.a = undefined;
        try {
            if (!someBitsSet(e, SETTLED)) e.settle(Error(DEINIT_ERROR));
            unsetBits(e, PENDING_REJECTION);
        } finally {
            try {
                finalize(e);
            } finally {
                try {
                    if (t) t.deinit();
                } finally {
                    var i = e.a;
                    e.a = undefined;
                    if (i) i.deinit();
                }
            }
        }
    }
};

Future.from = function from(e, t) {
    var i = new Future();
    i.settle(e, t);
    return i;
};

Future.fromError = function fromError(e) {
    validate(e, Boolean);
    return this.from(e);
};

Future.fromResult = function fromResult(e) {
    return isFuture(e) ? e : this.from(undefined, e);
};

Future.fromPromise = function fromPromise(e) {
    validate(e, isPromise);
    var t = new Future();
    e.then(t.settle.bind(t, undefined), t.settle.bind(t));
    return t;
};

Future.all = function all(e) {
    validate(e, isArray);
    var t = new Future();
    initAllJuncture(new AllJuncture(t, e));
    return t;
};

Future.race = function race(e) {
    validate(e, isArray);
    var t = new Future();
    initRaceJuncture(new RaceJuncture(t, e));
    return t;
};

Future.onUnhandledRejection = function onUnhandledRejection(e) {
    throw e.c;
};

Future.scheduler = new Scheduler(function finishPending(e) {
    e.finishPending();
});

function AllJuncture(e, t) {
    validateInstance(this, AllJuncture);
    this.R = e;
    this.N = t.slice();
    this.u = 0;
}

AllJuncture.prototype.deinit = function deinit() {
    this.R = undefined;
    var e = this.N;
    this.N = undefined;
    this.u = undefined;
    forceDeinitFutures(e);
};

function initAllJuncture(e) {
    var t = e.N;
    for (var i = 0; i < t.length; i += 1) {
        var n = t[i];
        if (!isFuture(n)) continue;
        if (isBaseFuture(n)) {
            if (someBitsSet(n, CONSUMED)) throw Error(CONSUMABLE_ERROR);
            if (someBitsSet(n, ERROR)) {
                settleAllJuncture(e, n);
                return;
            }
            if (someBitsSet(n, SUCCESS)) {
                t[i] = quietlyConsume(n);
                continue;
            }
            if (!n.S) {
                setupFinalizer(n, settleAllJunctureAtIndex.bind(null, e, i));
                if (someBitsSet(n, PENDING)) e.u += 1;
                continue;
            }
        }
        t[i] = n.map(settleAllJunctureAtIndex.bind(null, e, i));
        e.u += 1;
    }
    if (e.u) {
        e.R.S = e.deinit.bind(e);
    } else {
        settleAllJuncture(e);
    }
}

function settleAllJunctureAtIndex(e, t, i, n) {
    var u = e.N;
    if (!u) return;
    e.u -= 1;
    if (i) {
        settleAllJuncture(e, i);
    } else {
        u[t] = n;
        if (!e.u) settleAllJuncture(e);
    }
}

function settleAllJuncture(e, t) {
    var i = e.R;
    e.R = undefined;
    var n = e.N;
    e.N = undefined;
    if (t) {
        i.settle(t);
        forceDeinitFutures(n);
    } else {
        i.settle(undefined, n);
    }
}

function RaceJuncture(e, t) {
    validateInstance(this, RaceJuncture);
    this.R = e;
    this.N = t.slice();
    this.u = 0;
}

RaceJuncture.prototype.deinit = function deinit() {
    this.R = undefined;
    var e = this.N;
    this.N = undefined;
    this.u = undefined;
    forceDeinitFutures(e);
};

function initRaceJuncture(e) {
    var t = e.N;
    var i = settleRaceJuncture.bind(null, e);
    for (var n = 0; n < t.length; n += 1) {
        var u = t[n];
        if (!isFuture(u)) {
            i(undefined, u);
            return;
        }
        if (isBaseFuture(u)) {
            if (someBitsSet(u, CONSUMED)) throw Error(CONSUMABLE_ERROR);
            if (someBitsSet(u, ERROR)) {
                i(u);
                return;
            }
            if (someBitsSet(u, SUCCESS)) {
                i(undefined, quietlyConsume(u));
                return;
            }
            if (!u.S) {
                setupFinalizer(u, i);
                if (someBitsSet(u, PENDING)) e.u += 1;
                continue;
            }
        }
        t[n] = u.map(i);
        e.u += 1;
    }
    if (e.u) {
        e.R.S = e.deinit.bind(e);
    } else {
        i();
    }
}

function settleRaceJuncture(e, t, i) {
    var n = e.R;
    e.R = undefined;
    var u = e.N;
    e.N = undefined;
    if (!n) return;
    n.settle(t, i);
    forceDeinitFutures(u);
}

function isDeinitError(e) {
    return isObject(e) && e instanceof Error && e.message === DEINIT_ERROR;
};

var CONSUMABLE_ERROR = "Expected a consumable future: one that has not been mapped or passed to .settle";

var DEINIT_ERROR = "DEINIT";

function alwaysThrow(e, t) {
    throw e || t;
}

function mapError(e, t, i) {
    return t ? e(t) : i;
}

function mapResult(e, t, i) {
    if (t) throw t;
    return e(i);
}

function mapFinally(e, t, i) {
    var n = e(t, i);
    if (isFuture(n)) {
        return n.mapResult(function resetResult() {
            return i;
        });
    }
    if (t) throw t;
    return i;
}

function mapFutureToPromise(e, t, i) {
    function settleFuturePromise(n, u) {
        unsetBits(e, PENDING_REJECTION);
        if (n) i(n);else t(u);
    }
    if (!e.S) setupFinalizer(e, settleFuturePromise);else e.map(settleFuturePromise);
}

function scheduleFuture(e) {
    (e.constructor.scheduler || Future.scheduler).push(e);
}

function setupPredecessorSuccessorPair(e, t) {
    e.l = t;
    t.a = e;
    if (someBitsSet(e, SETTLED)) scheduleFuture(e);
}

function setupFinalizer(e, t) {
    e.S = t;
    if (someBitsSet(e, SETTLED)) scheduleFuture(e);
}

function quietlyConsume(e) {
    setBits(e, CONSUMED);
    return e.c;
}

function finalize(e) {
    var t = e.S;
    e.S = undefined;
    if (isFunction(t)) {
        if (someBitsSet(e, ERROR)) t(e.c);else t(undefined, e.c);
    }
}

function isFunction(e) {
    return typeof e === "function";
}

function isArray(e) {
    return isObject(e) && e instanceof Array;
}

function isObject(e) {
    return e != null && typeof e === "object";
}

function isPromise(e) {
    return isObject(e) && isFunction(e.then);
}

function isBaseFuture(e) {
    return isObject(e) && e.constructor === Future;
}

function validate(e, t) {
    if (!t(e)) throw Error("Expected " + e + " to satisfy test " + t.name);
}

function validateInstance(e, t) {
    if (!(e instanceof t)) throw Error("Cannot call a class as a function");
}

function forceDeinitFutures(e) {
    if (isArray(e)) forceEach(e, deinitFuture);
}

function deinitFuture(e) {
    if (isFuture(e)) e.deinit();
}

function forceEach(e, t) {
    var i = undefined;
    for (var n = 0; n < e.length; n += 1) {
        var u = e[n];
        try {
            t(u);
        } catch (e) {
            i = e;
        }
    }
    if (i) throw i;
}

function someBitsSet(e, t) {
    return (e.f & t) !== 0;
}

function setBits(e, t) {
    e.f |= t;
}

function unsetBits(e, t) {
    e.f &= ~t;
}

function replaceStateBits(e, t) {
    e.f = e.f & UNSET_STATE | t;
}

function chooseAsapImplementation(e) {
    if (typeof self === "undefined" && typeof process === "object" && process && process.nextTick) {
        return process.nextTick;
    }
    if (typeof MessageChannel === "function") {
        var t = new MessageChannel();
        t.port1.onmessage = e;
        return function asap() {
            t.port2.postMessage(undefined);
        };
    }
    return setTimeout;
}